package jcsc;

import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Point;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;

import javax.swing.JPanel;

/**
 * F̑Ipl
 *
 *   @author   Hiroshi Okuda / Kys
 *   @version  0.0  26 May 2007
 */
public class JcscCmapPanel extends JPanel
	implements	MouseListener,
				MouseMotionListener {
	private final boolean DEBUG = false;
	
	private final int size = 240;

	private Jcsc jcsc;
	private ExtColor[] colordic = null;

	private int theta = 0;
	private int phi   = 56;
	private int ratio = 128;

	private int[] x;
	private int[] y;
	private int[] z;
	private int[] l;
	private int[] zidx = null;
	private int[] ctbl = null;

	private final int STATE_NONE = 0;
	private final int STATE_ROT  = 1;
	private final int STATE_ZOOM = 2;
	private int state = STATE_NONE;
	private int oldmx;
	private int oldmy;

	/**
	 * RXgN^
	 * @param parent
	 */
	public JcscCmapPanel(Jcsc parent) {
		jcsc = parent;

		setDoubleBuffered(true);
		setPreferredSize(new Dimension(size, size));

		addMouseListener(this);
		addMouseMotionListener(this);
	}

	/**
	 * J[Tv̂̃Zbg
	 * @param colordic
	 */
	public void setColordic(ExtColor[] colordic) {
		this.colordic = colordic;
	}

	/**
	 * `
	 */
	public void paint(Graphics g) {
		if (colordic == null) return;

		int w = getWidth();
		int h = getHeight();
		int hw = w / 2;
		int hh = h / 2;
		
		/*
		 * cose[u
		 */
		if (ctbl == null) {
			ctbl = new int[256];
			for(int i=0; i<256; i++){
				ctbl[i] = (int)(Math.cos(2.*Math.PI*i/256.)*512.5);
			}
		}

		/*
		 * WvZ
		 */
		Point p;
		ExtColor exc;
		if (zidx == null) {
			x = new int[colordic.length];
			y = new int[colordic.length];
			z = new int[colordic.length];
			l = new int[colordic.length];
			zidx = new int[colordic.length];
		}
		int i, imax, j, m;
		imax = colordic.length;

		for(i = 0; i < imax; i++){
			zidx[i] = i;

			exc = colordic[i];
			j = theta + exc.h;
			p = intRot(0,exc.s * exc.v / 255, j);
			m = p.y;
			p = intRot((exc.v * 2 - 255)
					-(exc.s * exc.v / 255)
							*(1600 + ctbl[(exc.h + 80) & 255]) / 1600,
					p.x, phi);
			j = 600 + p.y;
			x[i] = hw + m * j * ratio / 280000;
			y[i] = hh - p.x * j * ratio / 280000;
			z[i] = j;
			l[i] = j / 80;
		}

		for(i=0; i<imax; i++){
			for(j=0; j<i; j++){
				if(z[i-j] < z[i-j-1]){
					m = z[i-j];
					z[i-j] = z[i-j-1];
					z[i-j-1] = m;
					m = zidx[i-j];
					zidx[i-j] = zidx[i-j-1];
					zidx[i-j-1] = m;
				} else break;
			}
		}

		/*
		 * `
		 */
		g.setColor(getBackground());
		g.fillRect(0, 0, w, h);

		for(j=0; j<imax; j++){
			i = zidx[j];

			exc = colordic[i];
			g.setColor(exc.color);
			g.fillRect(x[i], y[i], l[i], l[i]);
		}
	}

	/**
	 * int]
	 * @param u
	 * @param v
	 * @param w
	 * @return
	 */
	public Point intRot(int u, int v, int w) {
		final Point p = new Point(0, 0);
		int s, c;

		if(w < 0) {
			w = 256 - (-w % 256);
		} else {
			w = w % 256;
		}
		s = ctbl[(w + 192) % 256];
		c = ctbl[w % 256];
		p.x = (u * c - v * s) / 512;
		p.y = (u * s + v * c) / 512;

		return p;
	}

	/**
	 * }EXXi[Q
	 *   @param e }EXCxg
	 */
	public void mouseMoved(MouseEvent e) {}
	public void mouseEntered(MouseEvent e) {}
	public void mouseExited(MouseEvent e) {}
	public void mouseClicked(MouseEvent e) {}

	/**
	 * hbOꂽ
	 * @param e
	 */
	public void mouseDragged(MouseEvent e) {
		e.consume();

		int dx = e.getX() - oldmx;
		int dy = e.getY() - oldmy;
		oldmx = e.getX();
		oldmy = e.getY();
		
		switch (state) {
			case STATE_NONE:
			break;
			
			// ]
			case STATE_ROT:			
			theta = (theta + 256 + dx) % 256;
			phi = (phi + 256 + dy) % 256;
			repaint();
			break;

			// gEk
			case STATE_ZOOM:
			ratio += -dy;
			ratio = Math.max(ratio,   64);
			ratio = Math.min(ratio, 1024);
			repaint();
			break;

			default:
			break;
		}			
	}

	/**
	 * {^ꂽ
	 */
	public void mousePressed(MouseEvent e) {
		e.consume();

		if (state == STATE_NONE) {
			// {^ꂽ
			if (e.getButton() == MouseEvent.BUTTON1) {
				if (DEBUG) System.out.println("STATE_ROT");

				state = STATE_ROT;
				oldmx = e.getX();
				oldmy = e.getY();

				// F|CgĂ΂̐FI
				for (int j = colordic.length - 1; j > 0; j--) {
					int i = zidx[j];
					if (
						(oldmx >= x[i]) && (oldmx < x[i] + l[i]) &&
						(oldmy >= y[i]) && (oldmy < y[i] + l[i])
						)
					{
						jcsc.setDefaultColor(colordic[i].color);
						break;
					}
				}

			} else
			// E{^ꂽ
			if (e.getButton() == MouseEvent.BUTTON3) {
				if (DEBUG) System.out.println("STATE_ZOOM");

				state = STATE_ZOOM;
				oldmx = e.getX();
				oldmy = e.getY();

			}
		}
	}

	/**
	 * {^ꂽ
	 */
	public void mouseReleased(MouseEvent e) {
		state = STATE_NONE;

		if (DEBUG) System.out.println("STATE_NONE");
	}

}
