package jte.editpanel;

import java.awt.Color;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.event.MouseWheelEvent;
import java.awt.event.MouseWheelListener;

import javax.swing.JPanel;
import javax.swing.border.EtchedBorder;

import jte.Common;
import jte.PropertyManage;
import tangram.PieceInfo;
import tangram.PieceInfoExt;
import tangram.Tangram;
import tangram.svcs.Imgs;
import tangram.svcs.Listios;
import tangram.svcs.Opes;

/**
 * ҏWplNX
 *
 *   @author   Hiroshi Okuda / Kys
 *   @version  0.1  14 Apr.2007
 */
public class EditPanel
		extends JPanel
		implements  MouseListener,
					MouseMotionListener,
					MouseWheelListener
{
	private final boolean DEBUG = false;

	private Tangram tan;
	private Listios lios;

	private Imgs imgs;
	private Opes opes;

	/**
	 * RXgN^
	 * @param resm
	 * @param propm
	 * @param cmn
	 */
	public EditPanel() {
		setBorder(new EtchedBorder());

		tan = Common.getTangram();
		lios = Common.getListios();

		imgs = new Imgs(PropertyManage.editImageRatio);
		opes = new Opes(tan, imgs);
		loadConfigurations();

		addMouseMotionListener(this);
		addMouseListener(this);
		addMouseWheelListener(this);

		resetColors();
	}

	/**
	 * vpeB[ݒ[h
	 */
	public void loadConfigurations() {
		opes.setVertexAttach(PropertyManage.vertexAttach != 0);
		opes.setAdaptiveRotation(PropertyManage.adaptiveRotation != 0);
	}

	/**
	 * ^O`T[rXԂ
	 * @return
	 */
	public Imgs getImgs() {
		return imgs;
	}
	
	/**
	 * g嗦Ԃ
	 * @return
	 */
	public double getRatio() {
		return imgs.ratio;
	}

	/**
	 * g嗦̃Zbg
	 * @return
	 */
	public void setRatio(double ratio) {
		imgs.setRatio(ratio);

		PropertyManage.editImageRatio = ratio;		
	}

	/**
	 * F̃Zbg
	 */
	public void resetColors() {
		imgs.setPiecesColor(PropertyManage.piecesColor);
		imgs.setPiecesColord(PropertyManage.piecesColord);
	}

	/**
	 * `p^OZbg
	 * @param tan
	 */
	public void setTangram(Tangram tan) {
		if (!(this.tan.geombank.equals(tan.geombank))) {
			this.tan = new Tangram(tan.geombank, tan.pinit);
		}
		this.tan.importPinfo(tan.pinfo);
		this.tan.piext = tan.piext;
	}

	/**
	 * `p^OZbg
	 * @param pi
	 */
	public void setTangram(PieceInfo pinfo) {
		this.tan.importPinfo(pinfo);
	}

	/**
	 * ʂ`
	 * GuideImageQuality:
	 *   IMAGEQUALITY_NORMALcpX`AAȂ
	 *   IMAGEQUALITY_BETTERcpX`AA
	 *   IMAGEQUALITY_BEST  c@@@@V
	 *   @param g `
	 */
	private int bufWidth = 0;
	private int bufHeight = 0;
	public void paint(Graphics g) {
		Graphics2D g2 = (Graphics2D)g;

		// TCY
		if (getWidth() != bufWidth ||
			getHeight() != bufHeight) {
				if (DEBUG) System.out.println("Resize EditPnael");				

				bufWidth = getWidth();
				bufHeight = getHeight();

				imgs.setSlide(bufWidth/2, bufHeight/2);

		}

		g2.setColor(PropertyManage.backgroundColor);
		g2.fillRect(0, 0, bufWidth, bufHeight);

		// AAݒ
		imgs.setAntiAliasing
			(PropertyManage.editImageQuality > PropertyManage.IMAGEQUALITY_NORMAL);

		if (Common.isTabled()) {
			// e[u\
			drawTable(g2);
			return;
		}

		// X[vs[X`
		imgs.drawPieces(g2, tan.geombank, tan.pinfo, tan.piext,
						PieceInfoExt.MODE_SLEEP, true,
						false, true, null);

		// ANeBus[X`
		imgs.drawPieces(g2, tan.geombank, tan.pinfo, tan.piext,
						PieceInfoExt.MODE_ACTIVE, false,
						false, true, null);

		// I``
		if (opes.getStateOpe() == Opes.STATE_OPE_SLCTRECT) {
			imgs.drawSelectRect(g2, opes.rxs, opes.rys, opes.rxe, opes.rye);
		}

		// g嗦\
		if (opes.getStateOpe() == Opes.STATE_OPE_ZOOMING) {
			g2.setColor(PropertyManage.piecesColor);
			g2.drawString(((int)(imgs.ratio*100)) + "%", 2, 12);
		}
	}

	/**
	 * e[u\
	 * @param g2
	 */
	private final int tblpad = 10;
	private int tblcols;
	private int tblrows;
	private void drawTable(Graphics2D g2) {
		if (DEBUG) System.out.println("Draw Table");				

		double oldratio = imgs.ratio;
		Color color;
		Font font;

		tblcols = (bufWidth - tblpad*2) / 90;
		tblrows = (bufHeight - tblpad*2) / 90;

		double w = (double)(bufWidth - tblpad*2) / (double)tblcols;
		double h = (double)(bufHeight - tblpad*2) / (double)tblrows;

		double r = ((w < h) ? w : h) / 200 * 0.25 * imgs.ratio;
		imgs.setRatio(r);
		font = new Font(Font.SERIF, Font.PLAIN, 11);
		color = new Color(
			255 - PropertyManage.backgroundColor.getRed(),
			255 - PropertyManage.backgroundColor.getGreen(),
			255 - PropertyManage.backgroundColor.getBlue(),
			255);
		PieceInfo pi;
		for (int y = 0; y < tblrows; y++) {
			for (int x = 0; x < tblcols; x++) {
				int idx = Common.getListIndex()
				          - tblrows*tblcols/2
				          + y*tblcols+x;
				if (tblrows%2 == 0) idx += (tblcols+1)/2;
				if (idx < 0 || idx >= lios.getCount()) continue;
				imgs.setSlide(tblpad+(int)(w*((double)x+0.5)),
				              tblpad+(int)(h*((double)y+0.5)));
				pi = lios.get(idx);
				imgs.drawPieces(g2, tan.geombank, pi, tan.piext,
								PieceInfoExt.MODE_ALL, true,
								false, true, null);				
				g2.setColor(color);
				g2.setFont(font);
				String s = pi.name.toString();
				if (s.length() > 12) {
					s = s.substring(0, 12) + "...";
				}
				g2.drawString(s,
					tblpad+(int)(w*(double)x)+20,
					tblpad+(int)(h*(double)(y+1))-7);
			}
		}

		// n
		imgs.setRatio(oldratio);		
		imgs.setSlide(bufWidth/2, bufHeight/2);
	}

	/**
	 * }EXXi[Q
	 *   @param e }EXCxg
	 */
	public void mouseDragged(MouseEvent e) {
		e.consume();

		if (!Common.isTabled()) {
			opes.mouseDragged(e);
			repaint();
		}
	}

	/**
	 * {^ꂽ
	 *   @param e }EXCxg
	 */
	public void mousePressed(MouseEvent e) {
		e.consume();

		if (!Common.isTabled()) {
			opes.mousePressed(e);
			repaint();
		} else {
			
			// e[u\ȂAe[ȗI
			int idx;
			int x = e.getX();
			int y = e.getY();
			if (x <= tblpad || x > bufWidth -tblpad ||
			    y <= tblpad || y > bufHeight-tblpad) {
			    return;
			} else {
				idx  = Common.getListIndex()
					 +  (x-tblpad)*tblcols/(bufWidth -tblpad*2)
					 + ((y-tblpad)*tblrows/(bufHeight-tblpad*2))*tblcols
				     + ((tblrows%2 == 0) ? (tblcols+1)/2 : 0)
					 - tblrows*tblcols/2;
				Common.setListIndex(idx);
			}
		}
	}
	
	/**
	 * {^ꂽ
	 *   @param e }EXCxg
	 */
	public void mouseReleased(MouseEvent e) {
		e.consume();

		if (!Common.isTabled()) {
			int laststate = opes.getStateOpe();
			opes.mouseReleased(e);
			repaint();

			// s[XŒ莞AAhD[obt@ɃGg[
			if (laststate == Opes.STATE_OPE_WAIT &&
				opes.getStateOpe() == Opes.STATE_OPE_NONE) {
				Common.actionEditEntry();
			}

			// KChpl`
			if ((laststate == Opes.STATE_OPE_WAIT &&
				opes.getStateOpe() == Opes.STATE_OPE_NONE) ||
				opes.isChangedRatio()) {
				PropertyManage.editImageRatio = getRatio();
				Common.refreshGuidePanel(tan, false);
			}
		}
	}
	
	/**
	 * }EXړ
	 *   @param e }EXCxg
	 */
	public void mouseMoved(MouseEvent e) {
		e.consume();

		if (!Common.isTabled()) {
			opes.mouseMoved(e);

			if (opes.getStateOpe() != Opes.STATE_OPE_NONE) {
				repaint();
			}
		}
	}

	public void mouseWheelMoved(MouseWheelEvent e) {}
	public void mouseEntered(MouseEvent e) {}
	public void mouseExited(MouseEvent e) {}
	public void mouseClicked(MouseEvent e) {}

}
