package jte.guidepanel;

import java.awt.BasicStroke;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Image;
import java.awt.RenderingHints;
import java.awt.Stroke;
import java.awt.image.BufferedImage;
import java.awt.image.ImageObserver;
import java.awt.image.ImageProducer;
import java.awt.image.MemoryImageSource;
import java.awt.image.PixelGrabber;

import javax.swing.JPanel;

import jte.Common;
import jte.PropertyManage;
import tangram.PieceInfo;
import tangram.PieceInfoExt;
import tangram.Tangram;
import tangram.svcs.Imgs;

/**
 * KChpl
 *
 *   @author   Hiroshi Okuda / Kys
 *   @version  0.3  26 May 2007
 */
public class GuideImagePanel extends JPanel {
	private final boolean DEBUG = false;

	private Tangram tan;
	private Imgs imgs;

	private GuideToolBar gtb;

	private int bufWidth = 0;
	private int bufHeight = 0;
	private int[] gamma = null;
	private int[] invgamma = null;
	private BufferedImage bufimg = null;
	private Image bbufimg;
	private Graphics bufg;
	private PixelGrabber pg;	
	private int[] pixels;
	private int[] bpixels;
	private ImageProducer ip;

	private final Stroke gpLineStrokex1 =
		new BasicStroke(1.0f, BasicStroke.CAP_BUTT, BasicStroke.JOIN_BEVEL);
	private final Stroke gpLineStrokex4 =
		new BasicStroke(4.0f, BasicStroke.CAP_BUTT, BasicStroke.JOIN_BEVEL);

	/**
	 * RXgN^
	 * @param resm
	 * @param propm
	 * @param cmn
	 */
	public GuideImagePanel(GuideToolBar gtb) {
		imgs = new Imgs();
		tan = new Tangram();

		this.gtb = gtb;

		setPreferredSize(new Dimension(128,128));
		resetColors();
	}

	/**
	 * F̃Zbg
	 */
	public void resetColors() {
		imgs.setPiecesColor(PropertyManage.piecesColor);
		imgs.setPiecesColord(PropertyManage.piecesColord);
	}

	/**
	 * `p^OZbg
	 * @param tan
	 */
	public void setTangram(Tangram tan) {
		if (!(this.tan.geombank.equals(tan.geombank))) {
			this.tan = new Tangram(tan.geombank, tan.pinit);
		}
		this.tan.importPinfo(tan.pinfo);
		this.tan.piext = tan.piext;
	}

	/**
	 * `p^OZbg
	 * @param pi
	 */
	public void setTangram(PieceInfo pinfo) {
		this.tan.importPinfo(pinfo);
	}

	/**
	 * KChplɕ`
	 * GuideImageQuality:
	 *   IMAGEQUALITY_NORMALcpX`AAȂ
	 *   IMAGEQUALITY_BETTERcpX`AA
	 *   IMAGEQUALITY_BEST  cTvOϊ胊
	 */
	public void paint(Graphics g) {
		if (DEBUG) System.out.println("GuideImagePanel paint");
		double ratio = Common.getEditImageRatio();

		if (PropertyManage.guideImageQuality <= PropertyManage.IMAGEQUALITY_BETTER) {

			// 掿uǂvɂȂƂɃobt@Zbg邽߂
			// bufimgnullɂ
			bufimg = null;

			// {ŃpX`
			bufWidth = getWidth();
			bufHeight = getHeight();
			Graphics2D g2 = (Graphics2D)g;

			g2.setColor(PropertyManage.backgroundColor);
			g2.fillRect(0, 0, bufWidth, bufHeight);
			imgs.setRatio(ratio * PropertyManage.guideImageRatio);

			// AAݒ
			imgs.setAntiAliasing
				(PropertyManage.guideImageQuality == PropertyManage.IMAGEQUALITY_BETTER);

			g2.setRenderingHint(RenderingHints.KEY_INTERPOLATION,
			RenderingHints.VALUE_INTERPOLATION_BICUBIC);
			imgs.setSlide(bufWidth/2, bufHeight/2);
			imgs.drawPieces(g2, tan.geombank, tan.pinfo, tan.piext,
							PieceInfoExt.MODE_ALL,
							!gtb.isUnfilling(),
							false, true,
							(gtb.isUnfilling()) ? gpLineStrokex1 : null);
		} else {
			// lTvO
			if (gamma == null || invgamma == null) {
				gamma = new int[256];
				invgamma = new int[256*16];
				double gconst = PropertyManage.monitorGamma;

				// e[u
				for (int n = 0; n < 256; n++) {
					gamma[n] = (int)
						(Math.pow((double)n/255.0, gconst)*255.0);
				}
				// te[u
				for (int n = 0; n < 256*16; n++) {
					invgamma[n] = (int)
						(Math.pow((double)n/4095.0, 1/gconst)*255.0);
				}
			}

			// `TCY̕ύXA掿uʁEǂv
			// uǂvɂȂۂɁAobt@Zbg
			if (bufimg == null ||
				bufWidth != getWidth() ||
				bufHeight != getHeight()
				) {

				// obt@
				if (DEBUG) System.out.println("Resize GuideImagePnael");				
				bufWidth = getWidth();
				bufHeight = getHeight();

				if (bufimg != null) bufimg.flush();
				bufimg = new BufferedImage(bufWidth*4, bufHeight*4, BufferedImage.TYPE_INT_ARGB);
				bufg = bufimg.getGraphics();

				pixels = new int[bufHeight * bufWidth * 16];
				bpixels = new int[bufHeight * bufWidth];
				ip = new MemoryImageSource(
						bufWidth, bufHeight, bpixels, 0, bufWidth);
			}
				
			Graphics2D g2 = (Graphics2D)g;
			Graphics2D bg = (Graphics2D)bufg;
	

			bg.setColor(PropertyManage.backgroundColor);
			bg.fillRect(0, 0, bufWidth*4, bufHeight*4);
			imgs.setRatio(ratio * 4*PropertyManage.guideImageRatio);
			imgs.setAntiAliasing(false);
			imgs.setSlide(bufWidth*2, bufHeight*2);
			imgs.drawPieces(bg, tan.geombank, tan.pinfo, tan.piext,
							PieceInfoExt.MODE_ALL,
							!gtb.isUnfilling(),
							false, false,
							(gtb.isUnfilling()) ? gpLineStrokex4 : null);
		
			pg = new PixelGrabber(bufimg,
					0, 0, bufWidth*4, bufHeight*4, pixels, 0, bufWidth*4);
	        try {
	            pg.grabPixels();
	        } catch (InterruptedException e) {
	                    System.err.println("interrupted waiting for pixels!");
	                    return;
	        }
	        if ((pg.getStatus() & ImageObserver.ABORT) != 0) {
	                    System.err.println("image fetch aborted or errored");
	                    return;
	        }

			int iy4 = 0;
			for (int iy = 0; iy < bufHeight; iy++, iy4+=4){
				int ix4 = 0;
				for (int ix = 0; ix < bufWidth; ix++, ix4+=4){
					int vx, vr, vg, vb;
					vx = vr = vg = vb = 0;
					for (int jy = 0; jy < 4; jy++){
						for (int jx = 0; jx < 4; jx++){
							vx = pixels[ix4+jx+(iy4+jy)*bufWidth*4];
							vr += gamma[vx & 0x000000ff];
							vg += gamma[(vx>>8) & 0x000000ff];
							vb += gamma[(vx>>16) & 0x000000ff];
						}
					}
					bpixels[ix+iy*bufWidth] =
						invgamma[vr]+(invgamma[vg]<<8)+(invgamma[vb]<<16)
						| 0xff000000;
				}
			}
			bbufimg = createImage(ip);
	
			g2.drawImage(bbufimg, 0, 0, bufWidth, bufHeight, this);
		}
	}
}
