package jte.jtehtml;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Component;
import java.awt.Graphics2D;
import java.awt.RenderingHints;
import java.awt.Stroke;
import java.awt.geom.GeneralPath;
import java.awt.geom.Rectangle2D;
import java.awt.image.BufferedImage;
import java.awt.image.DataBufferInt;
import java.awt.image.ImageObserver;
import java.awt.image.PixelGrabber;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileOutputStream;
import java.io.OutputStreamWriter;
import java.util.Date;
import java.util.Locale;

import javax.imageio.ImageIO;
import javax.swing.JFileChooser;
import javax.swing.JOptionPane;
import javax.swing.JProgressBar;

import jte.PropertyManage;
import jte.ResourceManage;
import tangram.GeomCalc;
import tangram.GeomElem;
import tangram.PieceInfo;
import tangram.Tangram;
import tangram.svcs.Imgs;
import tangram.svcs.Listios;

/**
 * javade tangram HTMLo
 *
 *   @author   Hiroshi Okuda / Kys
 *   @version  0.1  14 Sep.2008
 */
public class JteHtmlMaker {
	private final boolean DEBUG = false;		// fobOp

	private String prefix;						// WtH_̃pX
	private String fullname;					// W̃t@C
	private String name;						// W̎t@C
	private String htmlprefix;					// HTML̃tH_̃pX
	private String htmlname;					// HTML̎t@C
	private final String pngextension = ".png";
	private final String htmlextension = ".html";
	private final String htmlfilesextension = "_files";
	private final String encodingName = "UTF-8";

	private int pxpadding_h = 10;				// Ԋu[px]
	private int pxmaxsize_h = 680-16;			// őTCY[px]
	private int maxsize = 150;					// VGbg̍őTCY[px]

	private int[] gamma = null;
	private int[] invgamma = null;
	
	private BufferedImage bufimg = null;
	private BufferedImage img = null;
	private Graphics2D bufg;
	private Graphics2D g;
	private PixelGrabber pg;	
	private int[] bufpixels;
	private int[] pixels;

	private final Stroke gpLineStrokex4 =
		new BasicStroke(4.0f, BasicStroke.CAP_BUTT, BasicStroke.JOIN_BEVEL);

	/**
	 * RXgN^
	 */
	public JteHtmlMaker() {}

	/**
	 * WHTML`i+摜jŏo
	 * @param tan
	 * @param lios
	 * @param imgs
	 * @param isAnswer
	 * @param progbar
	 * @return
	 */
	public boolean saveHTML(
		Tangram tan,
		Listios lios,
		Imgs imgs,
		boolean isAnswer,
		JProgressBar progbar
		){	

		int silcount;
		PieceInfo pi;
		
		if (DEBUG) System.out.println("JteHtmlMaker.saveHTML");
		if (DEBUG) {
			System.out.println(lios.getName());
			System.out.println(lios.getCount());
			System.out.println(tan.getPuzzleName());
			System.out.println(tan.getLocalPuzzleName(Locale.getDefault().getISO3Language()));
		}

		// 萔mF
		silcount = lios.getCount();
		if (silcount < 1) {
			if (DEBUG) System.out.println("> no silhouette...");
			return false;
		}

		// pXݒƃtH_쐬
		if (!setPrefix(lios.getName())) return false;

		// 摜TCY`FbN
		Rectangle2D[] silrectarray = new Rectangle2D[silcount];
		for (int silidx = 0; silidx < silcount; silidx++) {
			pi = lios.get(silidx);
			Rectangle2D r = null;

			// VGbg̃TCY߂
			for (int m = 0; m < pi.num; m++) {
				GeomElem ge = GeomCalc.getPointList(tan.geombank, pi, m, imgs.ratio, 0, 0);
				GeneralPath gp = imgs.getRawGeneralPath(ge);

				if (m == 0) {	r = gp.getBounds2D();
				} else {		r = r.createUnion(gp.getBounds2D());
				}
			}
			silrectarray[silidx] = r;
			if (DEBUG) {
				System.out.println((silidx+1)+":"+pi.name);
				System.out.println("("+r.getMinX()+","+r.getMinY()+")-("+r.getMaxX()+","+r.getMaxY()+")");
			}
		
			// őTCY𒴂ȂmF
			if (((r.getWidth()+8.0)/4 > maxsize)||((r.getHeight()+16)/4 > maxsize)) {
				if (DEBUG) System.out.println("> so learge ratio...");
				return false;
			}
		}

		// vOXo[\
		progbar.setVisible(true);

		// HTMLwb_̏o
		FileOutputStream fos;
		OutputStreamWriter osw;
		BufferedWriter bw;
		try {
			fos = new FileOutputStream(htmlprefix + htmlname + htmlextension);
			osw = new OutputStreamWriter(fos , encodingName);
			bw = new BufferedWriter(osw);
		} catch(Exception e) {
			System.out.println(e);
			return false;
		}

		StringBuilder sb = new StringBuilder();

		sb.append("<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01//EN\" \"http://www.w3.org/TR/html4/strict.dtd\">\n");
		sb.append("<html>\n\n");

		sb.append("<head>\n");
		sb.append("\t<meta http-equiv=\"Content-Type\" content=\"text/html;charset=");
		sb.append(encodingName);
		sb.append("\">\n");
		sb.append("\t<meta http-equiv=\"Content-Style-Type\" content=\"text/css\">\n");		
		sb.append("\t<title>");
		sb.append(tan.getLocalPuzzleName(Locale.getDefault().getISO3Language()));
		sb.append("</title>\n");
		sb.append("<style type=\"text/css\">\n<!--\n");
		sb.append(".mycaption {\n");
		sb.append("\tfont-size:x-small;\n");
		/*
		sb.append("\tfont-family:serif;\n");
		sb.append("\twhite-space:nowrap;\n");
		*/
		sb.append("\tpadding-left:");
		sb.append(pxpadding_h/2);
		sb.append("px; padding-right:");
		sb.append(pxpadding_h/2);
		sb.append("px\n");
		sb.append("}\n");
		sb.append("-->\n</style>\n");
		sb.append("</head>\n\n");

		sb.append("<body style=\"text-align:center;");
		sb.append("\n\tpadding-bottom:50px;");
		sb.append("\n\tcolor:#");
		Color fg = PropertyManage.piecesColor;
		if (fg.getRed() < 16) sb.append("0");
		sb.append(Integer.toHexString(fg.getRed()));
		if (fg.getGreen() < 16) sb.append("0");
		sb.append(Integer.toHexString(fg.getGreen()));
		if (fg.getBlue() < 16) sb.append("0");
		sb.append(Integer.toHexString(fg.getBlue()));

		sb.append(";\n\tbackground-color:#");
		Color bg = PropertyManage.backgroundColor;
		if (bg.getRed() < 16) sb.append("0");
		sb.append(Integer.toHexString(bg.getRed()));
		if (bg.getGreen() < 16) sb.append("0");
		sb.append(Integer.toHexString(bg.getGreen()));
		if (bg.getBlue() < 16) sb.append("0");
		sb.append(Integer.toHexString(bg.getBlue()));

		sb.append("\">\n");

		sb.append("<p style=\"font-size:xx-large\">");
		sb.append(fullname);
		sb.append("</p>\n");

		sb.append("<p style=\"font-size:medium\">");
		sb.append(tan.getLocalPuzzleName(Locale.getDefault().getISO3Language()));
		sb.append("</p>\n");

		sb.append("<p style=\"font-size:x-small\">\n");
		sb.append(prefix);
		sb.append("<br>");
		sb.append(new Date().toString());
		sb.append("</p>\n\n");

		try {
			bw.write(sb.toString());
		} catch(Exception e) {
			System.out.println(e);
			return false;
		}
		if (DEBUG) System.out.println(sb.toString());
		sb.delete(0, sb.length());

		// lTvO
		if (gamma == null || invgamma == null) {
			gamma = new int[256];
			invgamma = new int[256*16];
			double gconst = PropertyManage.monitorGamma;

			// e[u
			for (int n = 0; n < 256; n++) {
				gamma[n] = (int)
					(Math.pow((double)n/255.0, gconst)*255.0);
			}
			// te[u
			for (int n = 0; n < 256*16; n++) {
				invgamma[n] = (int)
					(Math.pow((double)n/4095.0, 1/gconst)*255.0);
			}
		}
		
		// {摜obt@̊m
		bufimg = new BufferedImage(maxsize*4, maxsize*4, BufferedImage.TYPE_INT_ARGB);
		bufpixels = new int[maxsize*4 * maxsize*4];
		bufg = (Graphics2D)bufimg.getGraphics();
		bufg.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_OFF);
		if (isAnswer) {	bufg.setStroke(gpLineStrokex4);
		} else {		bufg.setStroke(imgs.getSolidLineStroke());
		}
		
		// 摜o
		int imgcount = 0;
		for (int silidx = 0; silidx < silcount; ) {
			int silidxstop;
			int pos, pxsize_h;

			// 1sɂǂ܂ŋlߍނ߂
			pos = pxsize_h = pxpadding_h;
			for (silidxstop = silidx; silidxstop < silcount; pxsize_h = pos, silidxstop++) {
				pos += ((int)(silrectarray[silidxstop].getWidth()+16))/4 + pxpadding_h;
				if (pos > pxmaxsize_h) break;
			}

			// 1s`悷邽߂̃obt@m
			img = new BufferedImage(pxsize_h, maxsize, BufferedImage.TYPE_INT_ARGB);
			g = (Graphics2D)img.getGraphics();
			g.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_OFF);
			g.setColor(PropertyManage.backgroundColor);
			g.fillRect(0, 0, pxsize_h, maxsize);
			
			DataBufferInt dst = (DataBufferInt)img.getRaster().getDataBuffer();
			pixels = dst.getData();

			// HTML 1stableJn^Oimgo
			sb.append("<table style=\"margin-left:auto; margin-right:auto; width=\"");
			sb.append(pxsize_h);
			sb.append("\">\n");
			sb.append("\t<tr style=\"padding-top:0px; padding-bottom:10px\">\n");

			sb.append("\t\t<td colspan=\"");
			sb.append(silidxstop-silidx);
			sb.append("\" style=\"padding-top:0px; padding-bottom:0.5em\">");
			sb.append("<img src=\"");
			sb.append(htmlname);
			sb.append(htmlfilesextension);
			sb.append("/");
			sb.append(imgcount);
			sb.append(".png\" alt=\"silhouettes\" title=\"");
			sb.append(silidx+1);
			if (silidxstop > silidx+1) {
				sb.append("~");
				sb.append((silidxstop));
			}
			sb.append("\">\n");
			sb.append("\t</tr>\n");

			// HTMLVGbg^CgtrvfJn^O
			sb.append("\t<tr style=\"vertical-align:top\">\n");

			try {
				bw.write(sb.toString());
			} catch(Exception e) {
				System.out.println(e);
				return false;
			}
			if (DEBUG) System.out.println(sb.toString());
			sb.delete(0, sb.length());

			// 1s摜̍쐬
			pos = pxpadding_h;
			for (; silidx < silidxstop; silidx++) {

				// 1VGbg𓙔{ŕ`
				pi = lios.get(silidx);
				double sldx = -(silrectarray[silidx].getMinX()-8);
				double sldy = -(silrectarray[silidx].getMinY()-8);
				
				bufg.setColor(PropertyManage.backgroundColor);
				bufg.fillRect(0, 0, maxsize*4, maxsize*4);
				
				for (int m = 0; m < pi.num; m++) {
					if (pi.flip[m] >= 0) {	bufg.setColor(PropertyManage.piecesColor);
					} else {				bufg.setColor(PropertyManage.piecesColord);
					}
					GeomElem ge = GeomCalc.getPointList(tan.geombank, pi, m, imgs.ratio, sldx, sldy);
					GeneralPath gp = imgs.getRawGeneralPath(ge);
					if (isAnswer) {				bufg.draw(gp);
					} else {	bufg.fill(gp);	bufg.draw(gp);
					}
				}

				// {摜zɎ擾
				pg = new PixelGrabber(bufimg,
						0, 0, maxsize*4, maxsize*4, bufpixels, 0, maxsize*4);
		        try {
		            pg.grabPixels();
		        } catch (InterruptedException e) {
		                    System.err.println("interrupted waiting for pixels!");
		                    return false;
		        }
		        if ((pg.getStatus() & ImageObserver.ABORT) != 0) {
		                    System.err.println("image fetch aborted or errored");
		                    return false;
		        }

				// ؂蔲AoW̌vZ
				int w = ((int)(silrectarray[silidx].getWidth()+16))/4;
				int h = ((int)(silrectarray[silidx].getHeight()+16))/4;
				int dx1 = pos;
				int dy1 = maxsize - h;
				int dx2 = pos + w - 1;
				int dy2 = maxsize - 1;
				int sx1 = 0;
				int sy1 = 0;

				// kȂs摜zɏ
				int sy = sy1;
				for (int dy = dy1; dy <= dy2; dy++, sy+=4){
					int sx = sx1;
					for (int dx = dx1; dx <= dx2; dx++, sx+=4){
						int vx, vr, vg, vb;
						vx = vr = vg = vb = 0;
						for (int y = 0; y < 4; y++){
							for (int x = 0; x < 4; x++){
								vx = bufpixels[sx+x+(sy+y)*maxsize*4];
								vr += gamma[vx & 0x000000ff];
								vg += gamma[(vx>>8) & 0x000000ff];
								vb += gamma[(vx>>16) & 0x000000ff];
							}
						}
						pixels[dx+dy*pxsize_h] =
							invgamma[vr]+(invgamma[vg]<<8)+(invgamma[vb]<<16)
							| 0xff000000;
					}
				}

				// HTML VGbg̃^Cgo
				sb.append("\t\t<td class=\"mycaption\" width=\"");
				sb.append((int)((silrectarray[silidx].getWidth()+16)/4));
				sb.append("\">");
				
				sb.append("<i>");
				sb.append(silidx+1);
				sb.append(".<br></i>");
				sb.append(pi.name);
				
				sb.append("</td>\n");
				
				// ʒuXV
				pos += (silrectarray[silidx].getWidth()+16)/4 + pxpadding_h;

				// vOXo[XV
				progbar.setValue((int)(100*silidx/silcount));
			}

			// 1s摜png`ŏo
			try {
			    ImageIO.write(img, "png",
			    	new File(htmlprefix + htmlname + htmlfilesextension + "\\" + imgcount + pngextension));
			} catch (Exception e) {
			    e.printStackTrace();
			    return false;
			}

			// HTML 1stablevfI^O
			sb.append("\t</tr>\n</table>\n");

			imgcount++;
		}

		// HTMLtb^̏o
		sb.append("</body>\n");
		sb.append("</html>\n");

		try {
			bw.write(sb.toString());
			bw.close();
			osw.close();
			fos.close();
		} catch(Exception e) {
			System.out.println(e);
			return false;
		}
		if (DEBUG) System.out.println(sb.toString());

		// vOXo[
		progbar.setVisible(false);

		return true;
	}

	/**
	 * o͐肵AKvȃtH_Ă
	 * iprefixfullname,name̓e܂j
	 * @param filename
	 * @return
	 */
	private boolean setPrefix(String filename) {
		fullname = new File(filename).getName();
		name = new File(filename).getName();
		prefix =  new File(filename).getParent() + "\\";

		// jte̖Ogq𔲂Ă
		int extpos = name.lastIndexOf(".");
		if (extpos > 0) {
			name = name.substring(0,extpos);
		}

		String path = chooseFile(prefix + name + htmlextension);
		if (path == null) return false;
		if (new File(path).exists()) {
			// ㏑mF
			int ret = 
				JOptionPane.showConfirmDialog(
					(Component)null,
					ResourceManage.getString("message.fileoverwrite"),
					ResourceManage.getString("dialog.confirmation"),
					JOptionPane.INFORMATION_MESSAGE |
					JOptionPane.YES_NO_CANCEL_OPTION
				);
			switch (ret) {
			case JOptionPane.YES_OPTION:
				if (DEBUG) System.out.println("overwrite");
				break;
			default:
				if (DEBUG) System.out.println("cancel save as");
				return false;
			}
		}
		htmlname = new File(path).getName();
		
		// HTML̖Ogq𔲂Ă
		extpos = htmlname.lastIndexOf(".");
		if (extpos > 0) {
			htmlname = htmlname.substring(0,extpos);
		}

		htmlprefix =  new File(path).getParent() + "\\";

		// 摜tH_΍
		File filepath = new File(htmlprefix + htmlname + htmlfilesextension);
		if (filepath.exists()) {
			if (!filepath.isDirectory()) {
				if (DEBUG) System.out.println("> already "+filepath.getName()+" exists...");
				return false;
			}
		} else {
			if (DEBUG) System.out.println("> make "+filepath.getAbsolutePath());				
			if (!filepath.mkdir()) {
				if (DEBUG) System.out.println("> failed to make "+filepath.getName()+"...");
				return false;
			}
		}
		return true;
	}

	/**
	 * t@CI_CAO
	 * @param defaultpath nullȂI[v_CAOA
	 * ȊOdefaultpathŃZ[u_CAOJ
	 * @return _CAOŃLZꂽnullAȊO̓t@CpX
	 */
	private String chooseFile(String defaultpath) {
		String path;
		JFileChooser chooser = new JFileChooser();
		
		chooser.setCurrentDirectory(
			new File(defaultpath));
		
		int ret;
		if (defaultpath == null) {
			// I[v_CAO
			ret = chooser.showOpenDialog(null);
		} else {
			// Z[u_CAO
			chooser.setSelectedFile(new File(defaultpath));
			ret = chooser.showSaveDialog(null);
		}			

		if (ret == JFileChooser.APPROVE_OPTION) {
			path = chooser.getSelectedFile().getPath();
			if (DEBUG) System.out.println("path : " + path);

		} else {
			// _CAOŃLZꂽꍇ
			if (DEBUG) System.out.println("cancel dialog");
			path = null;
		}
		return path;
	}
	
}
