package somapp;

import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.Image;
import java.awt.Toolkit;
import java.awt.image.BufferedImage;
import java.awt.image.ImageObserver;
import java.awt.image.ImageProducer;
import java.awt.image.MemoryImageSource;
import java.awt.image.PixelGrabber;
import java.io.File;
import java.io.IOException;

import javax.imageio.ImageIO;

import tangram.PieceInfo;
import tangram.PieceInfoExt;
import tangram.PieceInfoTool;
import tangram.Tangram;
import tangram.svcs.Imgs;
import tangram.svcs.Listios;

/**
 * ^OVGbgSOMW[
 *   @(#)TSOM.class
 *
 *   @author   Hiroshi Okuda / Kys
 *   @version  0.0  11 Nov.2007
 */
public class TSOM {
	public boolean DEBUG = true;
	public boolean REFIMGOUT = false;
	public boolean MAPIMGOUT = false;
	public boolean MTXIMGOUT = false;

	private Tangram tan;
	private Listios lios;
	private Imgs imgs;

	static public final int DIMMENSION = 8; // m[hxNg̎
	private int MAPNODELEN = 100; // }bṽm[h
	private int SOITERATION = 25000; // wK
	private int SOSIGMA = 4; // ߖTwK̃m[h΍
	private int A0 = 0; // ̒ڎ1
	private int A1 = 1; // ̒ڎ2
	private NodeVector[] refnv;
	private NodeVector[] mapnv;

	private Image img = null;
	private final int W = 128;
	private final int H = 128;
	private final int SIGMA = 4; // W΍
	private final int SIGMA2 = SIGMA*2;
	private final double NRMTHR = 0.75; // K臒l
	private final double C = 1 / Math.pow(Math.sqrt(2 * Math.PI) * (double)SIGMA, 2);
	private double[] gv = null;
	private double thresh = 0;
	private double sum = 0;

	private int mapimgw = 400;
	private int mapimgh = 400;
	private int mapimgt = 20;
	public int mtximgw = 1200;
	public int mtximgh = 500;
	private int mtximgt = 20;
	public int mtximgcol = 32;
	private int mtximgrow = 5;
	public double mtximgr = 0.1;

	private String basedir = "../results/";
	
	/**
	 * RXgN^
	 */
	public TSOM() {
		tan = new Tangram();
		lios = new Listios();
		imgs = new Imgs();
	}

	/**
	 * NA
	 */
	public void clear() {
		refnv = null;
		mapnv = null;
	}
	
	/**
	 * 摜Ԃ
	 */
	public BufferedImage getImage() {
		return (BufferedImage)img;
	}

	/**
	 * W̓ǂݍ
	 */
	public boolean readSilhouettes(String res) {
		if (! lios.loadXML(tan, res, false, false)) {
			System.out.println("failed to read " + res + ".");
			return false;
		}
		System.out.println(lios.getName());
		System.out.println(lios.getCount() + " silhouettes");		
		if (lios.getCount() <= 0) return false;

		// m[hxNgm
		refnv = new NodeVector[lios.getCount()];
		for (int i = 0; i < refnv.length; i++) {
			refnv[i] = new NodeVector(DIMMENSION);
		}
		return true;
	}

	/**
	 * ȑgD}bv̍쐬
	 */
	public void precalculate() {
		if (refnv == null) return;

		// KEXze[u̍쐬
		gv = new double[SIGMA * 4];
		for (int r = 0; r < SIGMA * 4; r++) {
			gv[r] = C * Math.exp(
				-Math.pow((double)r, 2) / (2 * Math.pow((double)SIGMA, 2)));
		}

		// a`FbNA臒lvZ
		sum = 0;
		for (int v = -SIGMA2; v < SIGMA2; v++) {
			int rsq = v ^ 2;
			for (int u = -SIGMA2; u < SIGMA2; u++) {
				int r = (int)(Math.sqrt((double)(rsq + u ^ 2)) + 0.5);
				sum += gv[r];
			}
		}
		thresh = sum * NRMTHR;

		// Qƃm[hxNg
		for (int i = 0; i < refnv.length; i++) calcRefnv(i);

		// Qƃm[hxNg\
		if (DEBUG) {
			for (int i = 0; i < refnv.length; i++) {
				System.out.print("refference " + i + ": {");
				printnv(refnv[i]);
			}
		}

		// IWĩ}gNXo
		if (DEBUG) {
			int[] order = new int[refnv.length];
			for (int i = 0; i < refnv.length; i++) order[i] = i;
			String fn = basedir + "map_original.png";
			pngoutMatrix(fn, mtximgw, mtximgh, mtximgt, order,
							 mtximgcol, mtximgrow, mtximgr);
		}

		// Gxŏɕבւ
		System.out.print("complexicity order {");
		int[] order = new int[refnv.length];
		boolean[] check = new boolean[refnv.length];
		for (int i = 0; i < refnv.length; i++) check[i] = false;
		for (int n = 0; n < refnv.length; n++) {
			int c;
			for (c = 0; c < refnv.length; c++) {
				if (!check[c]) break;
			}
			for (int i = 0; i < refnv.length; i++) {
				if (check[i]) continue;
				if (refnv[i].data[0] < refnv[c].data[0]) {
					c = i;					
				}
			}
			check[c] = true;
			order[n] = c;

			if (DEBUG) {
				System.out.print(c + "  ");
			}
		}
		System.out.print("}");
		
		// Gxŕׂ}gNXo
		if (DEBUG) {
			String fn = basedir + "map_cmatrix.png";
			pngoutMatrix(fn, mtximgw, mtximgh, mtximgt, order,
							 mtximgcol, mtximgrow, mtximgr);
		}
	}

	public void calculate() {
		// Qƃm[hxNg㉺l߂
		NodeVector nv = new NodeVector(DIMMENSION);
		NodeVector nvmin = new NodeVector(DIMMENSION);
		NodeVector nvmax = new NodeVector(DIMMENSION);
		for (int j = 0; j < DIMMENSION; j++) {
			nvmax.data[j] = refnv[0].data[j];
			nvmin.data[j] = refnv[0].data[j];
			for (int i = 0; i < refnv.length; i++) {
				if (refnv[i].data[j] > nvmax.data[j]) nvmax.data[j] = refnv[i].data[j];
				if (refnv[i].data[j] < nvmin.data[j]) nvmin.data[j] = refnv[i].data[j];
			}
		}

		// }bvm[h̐
		MAPNODELEN = Math.max((int)(Math.sqrt((double)refnv.length / 50.0) * 100.0), 50);
		mapnv = new NodeVector[MAPNODELEN];
		for (int i = 0; i < MAPNODELEN; i++) {
			mapnv[i] = new NodeVector(DIMMENSION);
			for (int j = 0; j < DIMMENSION; j++) {
				double diff = nvmax.data[j] - nvmin.data[j];

				// _
				mapnv[i].data[j] = nvmin.data[j] + Math.random()*diff;

				// 
				//mapnv[i].data[j] = nvmin.data[j] - diff + 3 * diff * (double)i/ (double)MAPNODELEN;
			}
		}
		
		// }bvm[hxNg\
		if (DEBUG) {
			for (int i = 0; i < MAPNODELEN; i++) {
				System.out.print("initial map node " + i + ": {");
				printnv(mapnv[i]);
			}
			// }bvm[h̉摜o
			for (int i = 0 ; i < DIMMENSION-1; i++) {
				A0 = i;
				A1 = i+1;
				String fn = basedir + "map_ini" + i + ".png";
				pngoutMap(fn, 400, 400, 20, null);
			}
		}

		// wK
		int pitr = 0;
		System.out.print("Zo)");
		for (int itr = 0; itr < SOITERATION; itr++) {
			if (itr == pitr) {
				System.out.print("+");
				pitr += SOITERATION / 10;
			}
						
			// QƃxNgI
			nv = refnv[(int)(Math.random()*(double)refnv.length)];

			// Qƃm[hɍł߂}bvm[hT
			int c = 0;
			double rsqmin = -1;
			for (int i = 0; i < MAPNODELEN; i++) {
				double rsq = 0;
				for (int j = 0; j < DIMMENSION; j++) {
					rsq += Math.pow(mapnv[i].data[j] - nv.data[j], 2);
				}
				if (rsqmin < 0 || rsq < rsqmin) {
					c = i;
					rsqmin = rsq;
				}
			}

			// ߖT܂߂ĊwK
			double a = 1.0 - Math.pow((double)itr / (double)SOITERATION, 2);
			for (int i = 0; i < MAPNODELEN; i++) {				
				double h = a * Math.exp(
					-Math.pow((double)(i - c),2) /
					(2.0 * Math.pow(SOSIGMA, 2)));

				for (int j = 0; j < DIMMENSION; j++) {
					mapnv[i].data[j] +=
						h * (nv.data[j] - mapnv[i].data[j]);
				}
			}
		}
		System.out.println("<");

		// }bvm[hxNg\
		if (DEBUG) {
			for (int i = 0; i < MAPNODELEN; i++) {
				System.out.print("map node " + i + ": {");
				printnv(mapnv[i]);
			}
			// }bvm[h̉摜o
			for (int i = 0 ; i < DIMMENSION-1; i++) {
				A0 = i;
				A1 = i+1;
				String fn = basedir + "map" + i + ".png";
				pngoutMap(fn, 400, 400, 20, null);
			}
		}
		
		// }bvɃVGbgzu
		int[] mnn = new int[refnv.length];
		double[] smnn = new double[refnv.length];
		for (int n = 0; n < refnv.length; n++) {

			// Qƃm[hɍł߂}bvm[hT
			int c = 0;
			double rsqmin = -1;
			double[] rsq = new double[MAPNODELEN];
			for (int i = 0; i < MAPNODELEN; i++) {
				rsq[i] = 0;
				for (int j = 0; j < DIMMENSION; j++) {
					rsq[i] += Math.pow(mapnv[i].data[j] - refnv[n].data[j], 2);
				}
				if (rsqmin < 0 || rsq[i] < rsqmin) {
					c = i;
					rsqmin = rsq[i];
				}
			}

			// Oǂ̃m[hɂǂꂾ߂𒲂ׂ
			double rsqr;
			if (c == 0) {
				rsqr = rsq[1];
			} else if (c == MAPNODELEN-1) {
				rsqr = -rsq[MAPNODELEN-2];
			} else {					
				rsqr = (rsq[c-1] > rsq[c+1]) ? -rsq[c-1] : rsq[c+1];
			}

			mnn[n] = c;
			smnn[n] = rsqr / (rsq[c]+1);

			if (DEBUG) {
				System.out.println(n + ": " + mnn[n] + ", " + smnn[n]);
			}
		}

		// ɕבւ
		System.out.print("SOM order {");
		int[] order = new int[refnv.length];
		boolean[] check = new boolean[refnv.length];
		for (int i = 0; i < refnv.length; i++) check[i] = false;
		for (int n = 0; n < refnv.length; n++) {
			int c;
			for (c = 0; c < refnv.length; c++) {
				if (!check[c]) break;
			}
			for (int i = 0; i < refnv.length; i++) {
				if (check[i]) continue;
				if ((mnn[i] < mnn[c]) ||
					((mnn[i] == mnn[c]) && (smnn[i] < smnn[c]))) {
					c = i;					
				}
			}
			check[c] = true;
			order[n] = c;

			if (DEBUG) {
				System.out.print(c + "  ");
			}
		}
		System.out.print("}");

		// }bv̉摜o
		if (DEBUG) {
			for (int i = 0 ; i < DIMMENSION-1; i++) {
				A0 = i;
				A1 = i+1;
				String fn = basedir + "map_order" + i + ".png";
				pngoutMap(fn, mapimgw, mapimgh, mapimgt, order);
			}
		}
		
		// }gNXo
		if (DEBUG) {
			String fn = basedir + "map_matrix.png";
			pngoutMatrix(fn, mtximgw, mtximgh, mtximgt, order,
							 mtximgcol, mtximgrow, mtximgr);
		}
	}

	/**
	 * m[hxNg̕\
	 * @param nv
	 */
	public void printnv(NodeVector nv) {
		for (int i = 0; i < DIMMENSION; i++) {
			System.out.print(" " + nv.data[i]);
		}
		System.out.println(" }");
	}

	/**
	 * }bv摜PNGo
	 * @param fn
	 * @param w
	 * @param h
	 * @param trim
	 */
	public void pngoutMap(String fn, int w, int h, double trim, int[] order) {
		if (!MAPIMGOUT) return;
		
		double efw = (double)w - trim * 2;
		double efh = (double)h - trim * 2;
		double x, y;
		double xs, ys;
		double xe, ye;
	
		BufferedImage img;
		img = new BufferedImage(w, h, BufferedImage.TYPE_4BYTE_ABGR);
		Graphics2D g2 = (Graphics2D)img.getGraphics();
	
		// wih
		g2.setColor(new Color(228, 216, 220));
		g2.fillRect(0, 0, w, h);
	
		xs = xe = refnv[0].data[A0];
		ys = ye = refnv[0].data[A1];
		for (int n = 1; n < refnv.length; n++) {
			if (refnv[n].data[A0] < xs) xs = refnv[n].data[A0];
			if (refnv[n].data[A0] > xe) xe = refnv[n].data[A0];
			if (refnv[n].data[A1] < ys) ys = refnv[n].data[A1];
			if (refnv[n].data[A1] > ye) ye = refnv[n].data[A1];
		}
		for (int n = 0; n < refnv.length; n++) {
			x = (refnv[n].data[A0] - xs) / (xe - xs);
			y = (refnv[n].data[A1] - ys) / (ye - ys);
	
			imgs.setPiecesColor(new Color((int)(x*255), (int)((1-y)*128), 0, 128));
			imgs.setPiecesColord(new Color((int)(x*255), (int)((1-y)*128), 0, 128));
	
			x = trim + x * efw;
			y = trim + (1 - y) * efh;
	
			imgs.setRatio(0.1);
			imgs.setSlide((int)x, (int)y);
		
			lios.moveIndexTo(n);
			PieceInfo pi = new PieceInfo();
			PieceInfoTool.copy(pi, lios.get(n));
			imgs.drawPieces(g2, tan.geombank, pi, tan.piext, PieceInfoExt.MODE_ALL,
							true, false, true, null);
		}
	
		g2.setColor(new Color(255, 255, 255, 255));
		if (order != null) {
			// orderw肳ꂽꍇ͕בւ
			for (int n = 1; n < refnv.length; n++) {
				int ixs = (int)(trim + (refnv[order[n-1]].data[A0] - xs)/(xe - xs) * efw);
				int ixe = (int)(trim + (refnv[order[n  ]].data[A0] - xs)/(xe - xs) * efw);
				int iys = (int)(trim + (1 - (refnv[order[n-1]].data[A1] - ys)/(ye - ys)) * efh);
				int iye = (int)(trim + (1 - (refnv[order[n  ]].data[A1] - ys)/(ye - ys)) * efh);
				g2.drawLine(ixs, iys, ixe, iye);
				g2.fillRect(ixs-1, iys-1, 3, 3);
			}
		} else {
			// ordeȑꍇ̓}bvm[h
			for (int n = 1; n < MAPNODELEN; n++) {
				int ixs = (int)(trim + (mapnv[n-1].data[A0] - xs)/(xe - xs) * efw);
				int ixe = (int)(trim + (mapnv[n  ].data[A0] - xs)/(xe - xs) * efw);
				int iys = (int)(trim + (1 - (mapnv[n-1].data[A1] - ys)/(ye - ys)) * efh);
				int iye = (int)(trim + (1 - (mapnv[n  ].data[A1] - ys)/(ye - ys)) * efh);
				g2.drawLine(ixs, iys, ixe, iye);
				g2.fillRect(ixs-1, iys-1, 3, 3);
			}
		}
	
		// PNGt@C֏o
		if (DEBUG) System.out.println(" ->" + fn);
		File f = new File(fn);
		try {
			ImageIO.write(img, "png", f);
		} catch (IOException ex) {
			ex.printStackTrace();
		}
	}

	/**
	 * }gNX摜PNGo
	 * @param fn
	 * @param w
	 * @param h
	 * @param trim
	 * @param order
	 * @param col
	 * @param row
	 * @param r
	 */
	public void pngoutMatrix(
		String fn,
		int w,
		int h,
		double trim,
		int[] order,
		int col,
		int row,
		double r
	) {
		double efw = (double)w - trim * 2;
		double efh = (double)h - trim * 2;
		double x, y;
		int num = refnv.length;
		int ih = (int)trim*2 + ((num+col-1) / col) * (int)efh/row;
		BufferedImage img;
		img = new BufferedImage(w, ih ,	BufferedImage.TYPE_4BYTE_ABGR);
		Graphics2D g2 = (Graphics2D)img.getGraphics();
			
		g2.setColor(new Color(228, 216, 220));
		g2.fillRect(0, 0, w, ih);
		for (int n = 0; n < num; n++) {
			imgs.setPiecesColor(Color.BLACK);
			imgs.setPiecesColord(Color.BLACK);
	
			x = trim + ((double)(n % col) + 0.5) / col * efw;
			y = trim + ((double)(n / col) + 0.5) / row  * efh;
	
			imgs.setRatio(r);
			imgs.setSlide((int)x, (int)y);
		
			lios.moveIndexTo(order[n]);
			PieceInfo pi = new PieceInfo();
			PieceInfoTool.copy(pi, lios.get(order[n]));
			imgs.drawPieces(g2, tan.geombank, pi, tan.piext, PieceInfoExt.MODE_ALL,
							true, false, true, null);
		}
	
		// PNGt@C֏o
		if (DEBUG && MTXIMGOUT) {
			System.out.println(" ->" + fn);

			File f = new File(fn);
			try {
				ImageIO.write(img, "png", f);
			} catch (IOException ex) {
				ex.printStackTrace();
			}
		}

		this.img = img;
	}				

	/**
	 * nԖڂ̃VGbg̎Qƃm[hxNgvZ
	 * @param n
	 * @param imbuf
	 */
	public boolean calcRefnv(int n) {
		System.out.println("calcRefnv");

		NodeVector nv = new NodeVector(DIMMENSION);

		int[] px = new int[W*H];

		img = new BufferedImage(W, H, BufferedImage.TYPE_4BYTE_ABGR);
		Graphics2D g2 = (Graphics2D)img.getGraphics();
		g2.setColor(Color.WHITE);
		g2.fillRect(0, 0, W, H);
			
		// vZΏۂ̃VGbg`
		lios.moveIndexTo(n);
		PieceInfo pi = new PieceInfo();
		PieceInfoTool.copy(pi, lios.get(n));
		imgs.setRatio(0.20);
		imgs.setSlide(W/2, H/2);
		imgs.drawPieces(g2, tan.geombank, pi, tan.piext, PieceInfoExt.MODE_ALL,
						true, false, true, null);

		// fo
		PixelGrabber grabber = new PixelGrabber(img, 0, 0, W, H, px, 0, W);
		try {
			grabber.grabPixels();
		} catch(InterruptedException e) {
			e.printStackTrace();
			return false;
		}
		if ((grabber.status() & ImageObserver.ABORT) != 0) {
			System.out.println("image fetch aborted or errored.\n");
			return false;
		}

		// m[hxNgZo
		nv = calcRefnvBody(px);

		// |[go
		System.out.println("no." + lios.getIndex()
			+ "/" + lios.getCount() + " : " + pi.name);

		if (DEBUG && REFIMGOUT) {
			// 摜E`
			ImageProducer producer = new MemoryImageSource(W, H, px, 0, W);
			Image tmp = Toolkit.getDefaultToolkit().createImage(producer);
			g2.drawImage(tmp, 0, 0, W, H, null);
	
			String fn = basedir + "ref_" + lios.getIndex() + ".png";
			System.out.println(" ->" + fn);
			try {
				ImageIO.write((BufferedImage)img, "png", new File(fn));
			} catch (IOException ex) {
				ex.printStackTrace();
			}
		}

		refnv[n] = nv;
		return true;
	}

	/**
	 * ۂɉ摜m[hxNgZo
	 * @param px
	 * @return
	 */
	public NodeVector calcRefnvBody(int[] px) {
		NodeVector nv = new NodeVector(DIMMENSION);

		// 2l
		double[] pe = new double[W * H];
		boolean[] pj = new boolean[W * H];
		for (int i = 0; i < W * H; i++) {
			px[i] &= 0x80;
			px[i] >>= 7;
			pe[i] = 0;
			pj[i] = false;
		}

		// 摜8AGbWo
		int cur = 0;
		for (int y = 1; y < H - 1; y++) {
			cur = y * W + 1;
			for (int x = 1; x < W - 1; x++, cur++) {
				pj[cur] =
					(px[cur] == 0) &&
					((px[cur - 1] != 0) ||
					 (px[cur + 1] != 0) ||
					 (px[cur - W] != 0) ||
					 (px[cur + W] != 0));
			}
		}

		// ̃VGbg̕Gx
		cur = 0;
		double lco = 0;
		int ilco = 0;
		int So = 0;
		int xs = W;
		int xe = 0;
		int ys = H;
		int ye = 0;
		for (int y = 1; y < H - 1; y++) {
			cur = y * W + 1;
			for (int x = 1; x < W - 1; x++, cur++) {
				if (px[cur] != 0) continue;
				So += 1;
				if (pj[cur]) {
					ilco++;
					lco += calcEdgeElemLen(
						pj[cur - W - 1],
						pj[cur - W],
						pj[cur - W + 1],
						pj[cur - 1],
						pj[cur + 1],
						pj[cur + W - 1],
						pj[cur + W],
						pj[cur + W + 1]
					);
				}
				if (xs > x) xs = x;
				if (xe < x) xe = x;
				if (ys > y) ys = y;
				if (ye < y) ye = y;
			}
		}

		// 3x3̖̈ʐϗ
		int ww = Math.max(xe - xs, ye - ys) + 1;
		xs = (xe + xs - ww)/2;
		ys = (ye + ys - ww)/2;
		ww = (ww + 1)/3;
		for (int i = 2; i < 8; i++) nv.data[i] = 0;
		for (int row = 0; row < 3; row++) {
			for (int y = 0; y < ww; y++) {
				cur = (ys + ww*row + y)*W + xs;
				for (int col = 0; col < 3; col++) {
					for (int x = 0; x < ww; x++, cur++) {
						double v = (px[cur] != 0) ? 0.0 : 1.0;
						nv.data[2 + col] += v;
						nv.data[5 + row] += v;
					}
				}
			}
		}
		for (int i = 2; i < 8; i++) nv.data[i] /= So/100;
		
		// KEVAtB^
		double M = 0;
		for (int y = 0; y < H; y++) {
			cur = y * W;
			for (int x = 0; x < W; x++, cur++) {

				M = 0;
				int vs = y - SIGMA2;
				int ve = y + SIGMA2;
				if (vs < 0) vs = 0;
				if (ve > H) ve = H;
				for (int v = vs; v < ve; v++) {
					int rsq = (y - v) ^ 2;
					int us = x - SIGMA2;
					int ue = x + SIGMA2;
					if (us < 0) us = 0;
					if (ue > W) ue = W;
					int ref = v * W + us;
					for (int u = us; u < ue; u++, ref++) {
						int r = (int)(Math.sqrt((double)(rsq + (x - u) ^ 2)) + 0.5);
						M += (double)px[ref] * gv[r];
					}
				}
				pe[cur] = M;
				
			}
		}

		// gU摜8AGbWo
		cur = 0;
		for (int y = 1; y < H - 1; y++) {
			cur = y * W + 1;
			for (int x = 1; x < W - 1; x++, cur++) {
				pj[cur] =
					(pe[cur] < thresh) &&
					((pe[cur - 1] >= thresh) ||
					 (pe[cur + 1] >= thresh) ||
					 (pe[cur - W] >= thresh) ||
					 (pe[cur + W] >= thresh));
			}
		}

		// 摜HƁA̕Gx
		double lc = 0;
		double S = 0;
		int ilc = 0;
		cur = 0;
		for (int y = 0; y < H; y++) {
			for (int x = 0; x < W; x++, cur++) {
				if (x < SIGMA2 || x > W - SIGMA2 ||
					y < SIGMA2 || y > H - SIGMA2 ) {
						px[cur] = 128;
						continue;
					} 

				if (px[cur] != 0) px[cur] = 128;

				if (pe[cur] < thresh) {
					S += 1;
					if (pj[cur]) {
						ilc++;
	
						double lca = calcEdgeElemLen(
							pj[cur - W - 1],
							pj[cur - W],
							pj[cur - W + 1],
							pj[cur - 1],
							pj[cur + 1],
							pj[cur + W - 1],
							pj[cur + W],
							pj[cur + W + 1]
						);

						if (lca != 0) {
							px[cur] = 0xff0000;
							lc += lca;
						}
					}
				}
				if (px[cur] != 0) px[cur] |= (int)((1 - pe[cur] / sum) * 255) << 8;
			}
		}

		if (DEBUG && REFIMGOUT) {
			for (int i = 0; i < W * H; i++) px[i] += 0xff000000;
	
			// 3x3̈̐FtƖ_Ot
			for (int row = 0; row < 3; row++) {
				for (int y = 0; y < ww; y++) {
					cur = (ys + ww*row + y)*W + xs;
					for (int col = 0; col < 3; col++) {
						if (((row + col) & 1) == 0) {
							cur += ww;
							continue;
						}
						for (int x = 0; x < ww; x++, cur++) {
							px[cur] += (((px[cur] & 0xff) == 0) ? 0x14 : 0x8)*5;
						}
					}
				}
			}
			for (int j = 0; j < 3; j++) {
				for (int i = 0; i < (int)nv.data[2+j]/4; i++) {
					px[(ys + i)*W + xs + ww*j + ww/2] = 0xffffffff;
				}
				for (int i = 0; i < (int)nv.data[5+j]/4; i++) {
					px[(ys + ww*j + ww/2)*W + xs + i] = 0xffffffff;
				}
			}
		}
		
		if (DEBUG) System.out.println(" sigma = " + SIGMA);

		// ʕ\
		double cmplxo = lco * lco / So;
		double cmplx = lc * lc / S;
		if (DEBUG) {
			System.out.print(" Length l = " + lco + "(" + ilco + ")");
			System.out.println(" -> " + lc + "(" + ilc + ")"); 
			System.out.println(" Surface S = " + So + " -> " + S);
			System.out.println(" comlexity C=l^2/S = " + cmplxo + " -> " + cmplx);
			System.out.println(" comlexity diff. = " + (cmplxo - cmplx));
		}

		nv.data[0] = cmplxo;
		nv.data[1] = cmplx;
		return nv;
	}

	/**
	 * GbWfЂ̒␳
	 * @param d00
	 * @param d01
	 * @param d02
	 * @param d10
	 * @param d12
	 * @param d20
	 * @param d21
	 * @param d22
	 * @return
	 */
	double calcEdgeElemLen(
		boolean d00,boolean d01,boolean d02,
		boolean d10,            boolean d12,
		boolean d20,boolean d21,boolean d22)
	{
		if (
		(d00 ? 1 : 0) +
		(d01 ? 1 : 0) +
		(d02 ? 1 : 0) +
		(d10 ? 1 : 0) +
		(d12 ? 1 : 0) +
		(d20 ? 1 : 0) +
		(d21 ? 1 : 0) +
		(d22 ? 1 : 0) != 2
		) return 1.0;

		if (d00 || d02 || d20 || d22) {
			if (d01 || d10 || d12 || d21) return (Math.sqrt(5.0))/2.0;
			return Math.sqrt(2.0); 
		}
		return 1.0;
	}

}
