package tangram;

import java.util.*;

/**
 * ^ONX
 *   @(#)Tangram.class
 *
 *   @author   Hiroshi Okuda / Kys
 *   @version  0.0  25 Mar.2007
 */
public class Tangram {
	public GeomBank geombank; // }`oN
	public PieceInfo pinfo; // VGbg
	public PieceInfo pinit; // fBtHgVGbg
	public PieceInfoExt piext; // s[X̏
	public String puzzle; // pY̖O
	public LocalNameDict localnamedict; // pY̎

	private final String tangramname = "Tangram";

	/**
	 * RXgN^
	 *   ^Oŏ
	 */
	public Tangram() {
		setTraditionalTangram();
	}

	/**
	 * RXgN^
	 *   ^Oŏ
	 * @param tangramname
	 */
	public Tangram(String tangramname) {
		setTraditionalTangram();
	}

	/**
	 * ^OZbg
	 */
	public void setTraditionalTangram() {
		geombank = new GeomBank();
		geombank.setTraditionalTangram();
		pinit = new PieceInfo();
		pinfo = new PieceInfo();
		PieceInfoTool.setTraditionalTangram(pinit);
		PieceInfoTool.copy(pinfo, pinit);

		piext = new PieceInfoExt(pinit);

		this.puzzle = tangramname;

		localnamedict = new LocalNameDict();
		addLocalPuzzleName("jpn", "^O");

		breakup();
	}

	/**
	 * RXgN^
	 *   GeomBankPieceInfoŏ
	 *   @param gb GeomBank
	 *   @param pi CjVPieceInfo
	 */
	public Tangram(GeomBank gb, PieceInfo pi) {
		geombank = gb;
		pinit = new PieceInfo();
		pinfo = new PieceInfo();
		PieceInfoTool.copy(pinit, pi);
		PieceInfoTool.copy(pinfo, pi);

		piext = new PieceInfoExt(pinit);

		puzzle = "";

		localnamedict = new LocalNameDict();

		breakup();
	}

	/**
	 * pinit̃C|[g
	 */
	public void importPinit(PieceInfo pi) {
		PieceInfoTool.copy(pinit, pi);
		piext.init(pinit);
	}

	/**
	 * pinfõC|[g
	 */
	public void importPinfo(PieceInfo pi) {
		PieceInfoTool.copy(pinfo, pi);
	}

	/**
	 * pinfõGNX|[g
	 */
	public void exportPinfo(PieceInfo pi) {
		PieceInfoTool.copy(pi, pinfo);
	}

	/**
	 * pinfõGNX|[g
	 */
	public PieceInfo exportPinfo() {
		PieceInfo pi = new PieceInfo();
		PieceInfoTool.copy(pi, pinfo);
		return pi;
	}

	/**
	 * VGbg̃u[N
	 */
	public void breakup() {
		PieceInfoTool.copy(pinfo, pinit);
		for (int i = 0; i < pinfo.num; i++) {
			pinfo.x[i] *= 1.25;
			pinfo.y[i] *= 1.25;
			pinfo.x[i] += -10 + Math.random() * 20;
			pinfo.y[i] += -10 + Math.random() * 20;
			pinfo.rad[i] += -0.15 + Math.random() * 0.3;
		}

		piext.sleepAll();
	}

	/**
	 * pY̖Oݒ
	 *   @param s pY̖O
	 */
	public void setPuzzleName(String s) {
		puzzle = s.toString();
	}

	/**
	 * pY̖O擾
	 *   @return pY̖O
	 */
	public String getPuzzleName() {
		return puzzle.toString();
	}

	/**
	 * pY̎̃NA
	 */
	public void clearLocalPuzzleNames() {
		localnamedict.clear();
	}

	/**
	 * pY̎疼OԂ
	 * @param key
	 * @return
	 */
	public String getLocalPuzzleName(String key) {
		String name = localnamedict.getName(key);
		if (name.equals("")) {
			name = puzzle;
		}
		return name;
	}

	/**
	 * pỸXgԂ
	 * @return
	 */
	public List<LocalName> getLocalPuzzleNames() {
		return localnamedict.getList();
	}

	/**
	 * pY̎ɒǉ
	 * @param key
	 * @param name
	 */
	public void addLocalPuzzleName(String key, String name) {
		localnamedict.add(new LocalName(key, name));
	}

	/**
	 * VGbg̖Oݒ
	 *   @param s VGbg̖O
	 */
	public void setName(String s) {
		PieceInfoTool.setName(pinfo, s);
	}

	/**
	 * VGbg̖O擾
	 *   @return VGbg̖O
	 */
	public String getName() {
		return PieceInfoTool.getName(pinfo);
	}

	/**
	 * ԍñs[XX[vԂɂ
	 *   @param n s[Xԍ
	 */
	public void sleepOne(int n) {
		piext.sleepOne(n);
	}

	/**
	 * SẴs[XX[vԂɂ
	 */
	public void sleepAll() {
		piext.sleepAll();
	}

	/**
	 * ԍñs[XANeBuԂɂ
	 *   @param n s[Xԍ
	 */
	public void activeOne(int n) {
		piext.activeOne(n);
	}

	/**
	 * SẴs[XANeBuԂɂ
	 */
	public void activeAll() {
		piext.activeAll();
	}

	/**
	 * ANeBus[XQ̒SxW擾
	 *   @return ANeBus[XQ̒SxW
	 */
	public double getActivesOriginX() {
		int cnt = 0;
		double oxave = 0;

		for (int i = 0; i < pinfo.num; i++) {
			if (piext.mode[i] == PieceInfoExt.MODE_ACTIVE) {
				cnt++;
				oxave += pinfo.x[i];
			}
		}
		if (cnt > 0)
			oxave /= (double) cnt;
		return oxave;
	}

	/**
	 * ANeBus[XQ̒SyW擾
	 *   @return ANeBus[XQ̒SyW
	 */
	public double getActivesOriginY() {
		int cnt = 0;
		double oyave = 0;

		for (int i = 0; i < pinfo.num; i++) {
			if (piext.mode[i] == PieceInfoExt.MODE_ACTIVE) {
				cnt++;
				oyave += pinfo.y[i];
			}
		}
		if (cnt > 0)
			oyave /= (double) cnt;
		return oyave;
	}

	/**
	 * ANeBus[XQ̕ϊpx擾
	 *   @return ANeBus[XQ̕ϊpx
	 */
	public double getActivesRadian() {
		int cnt = 0;
		double radave = 0;

		for (int i = 0; i < pinfo.num; i++) {
			if (piext.mode[i] == PieceInfoExt.MODE_ACTIVE) {
				cnt++;
				radave += pinfo.rad[i];
			}
		}
		if (cnt > 0)
			radave = PieceInfoTool.redrad2pi(radave / (double) cnt);
		return radave;
	}

	/**
	 * ANeBus[X̑Ή]
	 *   @param drad ]px
	 */
	public void rotate(double drad) {
		double oxave = getActivesOriginX();
		double oyave = getActivesOriginY();
		double sin = Math.sin(drad);
		double cos = Math.cos(drad);

		for (int i = 0; i < pinfo.num; i++) {
			if (piext.mode[i] == PieceInfoExt.MODE_ACTIVE) {
				double dx = pinfo.x[i] - oxave;
				double dy = pinfo.y[i] - oyave;

				PieceInfoTool.rotate(pinfo, i, drad);
				PieceInfoTool.moveTo(
					pinfo,
					i,
					oxave + cos * dx - sin * dy,
					oyave + sin * dx + cos * dy);
			}
		}
	}

	/**
	 * ANeBus[X̉]
	 *   @param drad ]px
	 */
	public void rotateTo(double drad) {
		for (int i = 0; i < pinfo.num; i++) {
			if (piext.mode[i] == PieceInfoExt.MODE_ACTIVE)
				PieceInfoTool.rotateTo(pinfo, i, drad);
		}
	}

	/**
	 * ANeBus[X̋t]
	 */
	private void rotReverse() {
		for (int i = 0; i < pinfo.num; i++) {
			if (piext.mode[i] == PieceInfoExt.MODE_ACTIVE)
				PieceInfoTool.rotReverse(pinfo, i);
		}
	}

	/**
	 * ANeBus[X̊px15PʂɊۂ߂
	 *   @return ۂɉ]px
	 */
	public double adjustDegree15() {
		final int unitdeg = 15;

		// ANeBus[X0,1ȂȒPɍς܂
		int cnt = 0, last = 0;
		for (int i = 0; i < pinfo.num; i++) {
			if (piext.mode[i] == PieceInfoExt.MODE_ACTIVE) {
				cnt++;
				last = i;
			}
		}
		if (cnt == 0)
			return 0;
		if (cnt == 1) {
			return PieceInfoTool.adjustDegree(pinfo, last, unitdeg);
		}

		// ANeBus[XȂœKȊpxT
		double sumdrad = 0;
		double sumdradmin = -1;
		double drad = 0;
		double dradconf = 0;

		PieceInfo pi = new PieceInfo();
		for (int i = 0; i < pinfo.num; i++) {

			PieceInfoTool.copy(pi, pinfo);

			if (piext.mode[i] == PieceInfoExt.MODE_ACTIVE) {
				drad = PieceInfoTool.adjustDegree(pi, i, unitdeg);

				// ڃs[X̕␳pxɍ킹đŜ]ꍇA
				// eX̃s[XPʊpxǂꂭ炢Ă邩ςݏグ
				sumdrad = 0;
				for (int j = 0; j < pinfo.num; j++) {
					if (j != i && piext.mode[j] == PieceInfoExt.MODE_ACTIVE) {
						PieceInfoTool.rotate(pi, j, drad);
						sumdrad
							+= Math.abs(
								PieceInfoTool.adjustDegree(pi, j, unitdeg));
					}
				}
				// ̑ałȏꍇ̕␳pxɂ
				if (sumdradmin < 0 || sumdradmin > sumdrad) {
					sumdradmin = sumdrad;
					dradconf = drad;
				}
			}
		}
		rotate(dradconf); // ]
		return dradconf;
	}

	/**
	 * ANeBus[X̑Έړ
	 *   @param dx xړ
	 *   @param dy yړ
	 */
	public void move(double dx, double dy) {
		for (int i = 0; i < pinfo.num; i++) {
			if (piext.mode[i] == PieceInfoExt.MODE_ACTIVE)
				PieceInfoTool.move(pinfo, i, dx, dy);
		}
	}

	/**
	 * ANeBus[Ẍړ
	 *   @param x xW
	 *   @param y yW
	 */
	public void moveTo(double x, double y) {
		for (int i = 0; i < pinfo.num; i++) {
			if (piext.mode[i] == PieceInfoExt.MODE_ACTIVE)
				PieceInfoTool.moveTo(pinfo, i, x, y);
		}
	}

	/**
	 * ANeBus[X̐]
	 */
	public void flipH() {
		flip();
		rotReverse();

		double oxave = getActivesOriginX();

		for (int i = 0; i < pinfo.num; i++) {
			if (piext.mode[i] == PieceInfoExt.MODE_ACTIVE)
				pinfo.x[i] += 2 * (oxave - pinfo.x[i]);
		}
	}

	/**
	 * ANeBus[X̑Δ]
	 */
	private void flip() {
		for (int i = 0; i < pinfo.num; i++) {
			if (piext.mode[i] == PieceInfoExt.MODE_ACTIVE)
				PieceInfoTool.flip(pinfo, i);
		}
	}

	/**
	 * ANeBus[X̔]
	 */
	public void flipTo(double flp) {
		for (int i = 0; i < pinfo.num; i++) {
			if (piext.mode[i] == PieceInfoExt.MODE_ACTIVE)
				PieceInfoTool.flipTo(pinfo, i, flp);
		}
	}

}
